@@ -1,4 +1,4 @@
-#ident "@(#)tio.c	1.38 94/10/22 Copyright (c) 1993 Gert Doering"
+#ident "@(#)tio.c	4.1 97/01/12 Copyright (c) 1993 Gert Doering"
 
 /* tio.c
  *
@@ -8,13 +8,15 @@
 
 #include <unistd.h>
 #include <sys/types.h>
+#include <errno.h>
 
-#ifdef _AIX
-#include <sys/ioctl.h>
+#ifdef NeXT
+#include <sys/file.h>
 #endif
 
 #include "mgetty.h"
 #include "tio.h"
+
 #ifdef POSIX_TERMIOS
 static char tio_compilation_type[]="@(#)tio.c compiled with POSIX_TERMIOS";
 #endif
@@ -25,7 +27,7 @@ static char tio_compilation_type[]="@(#)tio.c compiled with SYSV_TERMIO";
 static char tio_compilation_type[]="@(#)tio.c compiled with BSD_SGTTY";
 #endif
 
-#ifdef SVR4
+#ifdef USE_TERMIOX
 # include <sys/termiox.h>
 #endif
 #ifdef _HPUX_SOURCE
@@ -112,6 +114,9 @@ static struct	speedtab {
 #ifdef	B57600
 	{ B57600, 57600, "57600" },
 #endif
+#ifdef	B76800
+	{ B76800, 76800, "76800" },
+#endif
 #ifdef	B115200
 	{ B115200,115200,"115200"},
 #endif
@@ -173,9 +178,9 @@ int tio_set _P2( (fd, t), int fd, TIO * t)	/*!! FIXME: flags, wait */
     if (t->c_cflag & CRTSCTS)
     {
         /* make sure RTS is asserted!!!!!! */
-        ioctl(STDIN, TIOCMGET, &modem_lines);
+        ioctl(fd, TIOCMGET, &modem_lines);
         modem_lines |= (TIOCM_RTS | TIOCM_DTR);
-        ioctl(STDIN, TIOCMSET, &modem_lines);
+        ioctl(fd, TIOCMSET, &modem_lines);
     }
 #endif /* sunos4 */
 #endif /* posix_termios */
@@ -204,18 +209,36 @@ int tio_check_speed _P1( (speed), int speed )
     return -1;
 }
 
-/* set speed, do not touch the other flags */
-int tio_set_speed _P2( (t, speed ), TIO *t, int speed )
+/* set speed, do not touch the other flags
+ * "speed" is given as numeric baud rate, not as Bxxx constant
+ */
+int tio_set_speed _P2( (t, speed ), TIO *t, unsigned int speed )
 {
+    int i, symspeed=0;
+
+    for ( i=0; speedtab[i].cbaud != 0; i++ )
+    {
+	if ( speedtab[i].nspeed == speed ) symspeed = speedtab[i].cbaud;
+    }
+
+    if ( symspeed == 0 )
+    {
+	errno=EINVAL;
+	lprintf( L_ERROR, "tss: unknown/unsupported bit rate: %d", speed );
+	return ERROR;
+    }
+
+    lprintf( L_NOISE, "tss: set speed to %d (%03o)", speed, symspeed );
+	
 #ifdef SYSV_TERMIO
-    t->c_cflag = ( t->c_cflag & ~CBAUD) | speed;
+    t->c_cflag = ( t->c_cflag & ~CBAUD) | symspeed;
 #endif
 #ifdef POSIX_TERMIOS
-    cfsetospeed( t, speed );
-    cfsetispeed( t, speed );
+    cfsetospeed( t, symspeed );
+    cfsetispeed( t, symspeed );
 #endif
 #ifdef BSD_SGTTY
-    t->sg_ispeed = t->sg_ospeed = B0;
+    t->sg_ispeed = t->sg_ospeed = symspeed;
 #endif
     return NOERROR;
 }
@@ -287,11 +310,7 @@ void tio_mode_sane _P2( (t, local), TIO * t, int local )
     t->c_iflag = BRKINT | IGNPAR | IXON | IXANY;
     t->c_oflag = OPOST | TAB3;
     /* be careful, only touch "known" flags */
-    t->c_cflag&= ~(CSIZE | CSTOPB | PARENB | PARODD | CLOCAL
-#ifdef LOBLK
-		   | LOBLK
-#endif
-		   );
+    t->c_cflag&= ~(CSIZE | CSTOPB | PARENB | PARODD | CLOCAL);
     t->c_cflag|= CS8 | CREAD | HUPCL | ( local? CLOCAL:0 );
     t->c_lflag = ECHOK | ECHOE | ECHO | ISIG | ICANON;
 
@@ -309,8 +328,9 @@ void tio_mode_sane _P2( (t, local), TIO * t, int local )
     t->c_cc[VSWTCH] = 0;
 #endif
 
-#else
+#else		/* BSD_SGTTY */
     t->sg_flags = ECHO | EVENP | ODDP;
+/*    t->sg_flags = ECHO; */
     t->sg_erase = 0x7f;            /* erase character */
     t->sg_kill  = 0x25;            /* kill character, ^u */
 #endif
@@ -386,8 +406,15 @@ void tio_map_cr _P2( (t, perform_mapping), TIO * t, int
 	t->c_iflag &= ~ICRNL;
 	t->c_oflag &= ~ONLCR;
     }
-#else
-#include "not implemented yet"
+#else		/* BSD_SGTTY (tested only on NeXT yet, but should work) */
+    if ( perform_mapping )
+    {
+        t->sg_flags |= CRMOD ;
+    }
+    else
+    {
+        t->sg_flags &= ~CRMOD ;
+    }
 #endif
 }
 
@@ -412,8 +439,8 @@ void tio_map_uclc _P2( (t, perform_mapping), TIO * t, int
 	t->c_oflag &= ~OLCUC;
 	t->c_lflag &= ~XCASE;
     }
-# else
-# include "not implemented yet"
+# else		/* BSD_SGTTY */
+#  include "not implemented yet"
 # endif
 #endif		/* BSDI */
 }
@@ -433,8 +460,15 @@ void tio_carrier _P2( (t, carrier_sensitive), TIO *t, int carrier_sensitive )
     {
 	t->c_cflag |= CLOCAL;
     }
-#else
-#include "not implemented yet"
+#else		/* BSD_SGTTY (tested only on NeXT yet, but should work) */
+    if ( carrier_sensitive )
+    {
+        t->sg_flags &= ~LNOHANG ;
+    }
+    else
+    {
+        t->sg_flags |= LNOHANG ;
+    }
 #endif
 }
 
@@ -487,7 +521,7 @@ void tio_carrier _P2( (t, carrier_sensitive), TIO *t, int carrier_sensitive )
 
 int tio_set_flow_control _P3( (fd, t, type), int fd, TIO * t, int type )
 {
-#ifdef SVR4
+#ifdef USE_TERMIOX
     struct termiox tix;
 #endif
 
@@ -504,14 +538,23 @@ int tio_set_flow_control _P3( (fd, t, type), int fd, TIO * t, int type )
 			t->c_cflag |= HARDWARE_HANDSHAKE;
     if ( type & FLOW_XON_IN )
 			t->c_iflag |= IXOFF;
+    /* for login, we want IXON|IXANY, for voice, we must not set IXANY! */
     if ( type & FLOW_XON_OUT )
-			t->c_iflag |= IXON | IXANY;
+    {
+	                t->c_iflag |= IXON;
+        if ( type & FLOW_XON_IXANY )
+	                t->c_iflag |= IXANY;
+    }
 #else
-#include "not yet implemented"
+# ifdef NEXTSGTTY
+    lprintf( L_WARN, "tio_set_flow_control: not yet implemented" );
+# else
+#  include "not yet implemented"
+# endif
 #endif
+
     /* SVR4 came up with a new method of setting h/w flow control */
-    /* unfortunately, it's broken in 4.2 and Solaris2! */
-#if defined(SVR4) && !defined(SVR42) &&!defined(solaris2)
+#ifdef USE_TERMIOX
     if (ioctl(fd, TCGETX, &tix) < 0)
     {
 	lprintf( L_ERROR, "ioctl TCGETX" ); return ERROR;
@@ -603,10 +646,12 @@ int tio_toggle_dtr _P2( (fd, msec_wait), int fd, int msec_wait )
      * Strange enough, on *some* platforms, you have to pass the mctl
      * flag word by value, on others by reference. Oh world...
      */
-#if defined(SVR4) && defined(TIOCMBIS)		/* SVR4 special */
+#if defined(TIOCMBIS) && \
+    ( defined(sun) || defined(SVR4) || defined(NeXT) || defined(linux) )
+    
     int mctl = TIOCM_DTR;
 
-#ifdef sun
+#if !defined( TIOCM_VALUE )
     if ( ioctl( fd, TIOCMBIC, &mctl ) < 0 )
 #else
     if ( ioctl( fd, TIOCMBIC, (char *) mctl ) < 0 )
@@ -615,7 +660,7 @@ int tio_toggle_dtr _P2( (fd, msec_wait), int fd, int msec_wait )
 	lprintf( L_ERROR, "TIOCMBIC failed" ); return ERROR;
     }
     delay( msec_wait );
-#ifdef sun
+#if !defined( TIOCM_VALUE)
     if ( ioctl( fd, TIOCMBIS, &mctl ) < 0 )
 #else
     if ( ioctl( fd, TIOCMBIS, (char *) mctl ) < 0 )
@@ -624,7 +669,7 @@ int tio_toggle_dtr _P2( (fd, msec_wait), int fd, int msec_wait )
 	lprintf( L_ERROR, "TIOCMBIS failed" ); return ERROR;
     }
     return NOERROR;
-#else						/* !SVR4 */
+#else						/* !TIOCMBI* */
 
     /* On HP/UX, lowering DTR by setting the port speed to B0 will
      * leave it there. So, do it via HP/UX's special ioctl()'s...
@@ -680,9 +725,9 @@ int tio_toggle_dtr _P2( (fd, msec_wait), int fd, int msec_wait )
     /* on SunOS, if you hangup via B0, the DTR line will *stay* low.
      * So: enable it manually again.
      */
-    ioctl(STDIN, TIOCMGET, &modem_lines);
+    ioctl(fd, TIOCMGET, &modem_lines);
     modem_lines |= (TIOCM_RTS | TIOCM_DTR);
-    ioctl(STDIN, TIOCMSET, &modem_lines);
+    ioctl(fd, TIOCMSET, &modem_lines);
 #endif
     result = tio_set( fd, &save_t );
     
@@ -773,3 +818,137 @@ int tio_flow _P2( (fd, restart_output), int fd, int restart_output )
 
     return r;
 }
+
+
+/* tio_drain(fd): wait for output queue to drain
+ */
+
+int tio_drain_output _P1( (fd), int fd )
+{
+#ifdef POSIX_TERMIOS
+    if ( tcdrain( fd ) == ERROR )
+    {
+	lprintf( L_ERROR, "tio_drain: tcdrain" ); return ERROR;
+    }
+#else
+# ifdef SYSV_TERMIO
+    if ( ioctl( fd, TCSBRK, 1 ) == ERROR )
+    {
+    	lprintf( L_ERROR, "tio_drain: TCSBRK/1" ); return ERROR;
+    }
+# else	/* no way to wait for data to drain with BSD_SGTTY */
+    lprintf( L_WARN, "tio_drain: expect spurious failures" );
+# endif
+#endif
+    return NOERROR;
+}
+
+/* send a BREAK signal to the tty
+ *
+ * kernel break via tcsendbreak() or ioctl(TCSENDBRK) lasts at least
+ * 0.25 seconds. We can speed up this by switching baud rate to 1/4,
+ * and then sending a 0-byte (activated #ifdef FAST_BREAK).
+ * 
+ * on some systems (don't we love it all?) the "real" break functions
+ * don't work, so we must use FAST_BREAK.
+ */
+
+int tio_break _P1((fd), int fd)
+{
+/* SunOS4 doesn't seem to like "tcsendbreak" (noop), so send a "long zero" */
+#if defined(sunos4) || defined(M_UNIX) || defined(FAST_BREAK)
+    TIO tio, tio_save; int speed; char null = 0;
+
+    lprintf( L_NOISE, "sending FAST break" );
+
+    if ( tio_get( fd, &tio ) == ERROR )
+    {
+	lprintf( L_ERROR, "tio_break: can't get tio" ); return ERROR;
+    }
+    tio_save = tio;
+    if ( (speed = tio_get_speed( &tio ) ) < 150 ||
+	 tio_set_speed( &tio, speed/4 ) == ERROR ||
+	 tio_set( fd, &tio ) == ERROR )
+    {
+	lprintf( L_ERROR, "tio_break: can't set 1/4 speed" ); return ERROR;
+    }
+    if ( write( fd, &null, 1 ) != 1 )
+    {
+	lprintf( L_ERROR, "tio_break: can't write 0-byte" ); return ERROR;
+    }
+
+    /* before we switch baud rates back, make sure zero byte has been sent!
+     */
+    if ( tio_drain_output( fd ) == ERROR ) return ERROR;
+
+    if ( tio_set( fd, &tio_save ) == ERROR )
+    {
+	lprintf( L_ERROR, "tio_break: can't reset old TIO state" ); return ERROR;
+    }
+
+#else	/* !FAST_BREAK -> use standard functions */
+
+    lprintf( L_NOISE, "sending system call break" );
+#ifdef POSIX_TERMIOS
+    if ( tcsendbreak( fd, 0 ) < 0 )
+    {
+    	lprintf( L_ERROR, "tcsendbreak() failed" );
+    	return ERROR;
+    }
+#endif
+#ifdef SYSV_TERMIO
+    if ( ioctl( fd, TCSBRK, 0 ) < 0 )
+    {
+    	lprintf( L_ERROR, "ioctl( TCSBRK ) failed" );
+    	return ERROR;
+    }
+#endif
+#ifdef BSD_SGTTY
+   if ( ioctl( fd, TIOCSBRK, 0 ) < 0 )
+   {
+   	lprintf( L_ERROR, "ioctl( TIOCSBRK ) failed" );
+   	return ERROR;
+   }
+   delay( 250 );
+   if ( ioctl( fd, TIOCCBRK, 0 ) < 0 )
+   {
+   	lprintf( L_ERROR, "ioctl( TIOCCBRK ) failed" );
+   	return ERROR;
+   }
+#endif
+#endif /* FAST_BREAK */
+   return NOERROR;
+}
+
+/* tio_get_rs232_lines()
+ *
+ * get the status of all RS232 status lines
+ * (coded by the TIO_F_* flags. On systems that have the BSD TIOCM_*
+ * flags, we use them, on others we may have to do some other tricks)
+ *
+ * "-1" can mean "error" or "not supported on this system" (e.g. SCO).
+ */
+
+int tio_get_rs232_lines _P1( (fd), int fd)
+{
+    int flags;
+#ifdef TIO_F_SYSTEM_DEFS
+    if ( ioctl(fd, TIOCMGET, &flags ) < 0 )
+    	lprintf( L_ERROR, "tio_get_rs232_lines: TIOCMGET failed" );
+
+#else /* !TIO_F_SYSTEM_DEFS */
+    flags=-1;
+#endif
+
+    if ( flags != -1 )
+    {
+        lprintf( L_NOISE, "tio_get_rs232_lines: status:" );
+        if ( flags & TIO_F_RTS ) lputs( L_NOISE, " RTS" );
+        if ( flags & TIO_F_CTS ) lputs( L_NOISE, " CTS" );
+        if ( flags & TIO_F_DSR ) lputs( L_NOISE, " DSR" );
+        if ( flags & TIO_F_DTR ) lputs( L_NOISE, " DTR" );
+        if ( flags & TIO_F_DCD ) lputs( L_NOISE, " DCD" );
+        if ( flags & TIO_F_RI  ) lputs( L_NOISE, " RI" );
+    }
+    return flags;
+}
